@@ -79,6 +79,7 @@ struct _BackgroundClass
 static XdpImplAccess *access_impl;
 static XdpImplBackground *background_impl;
 static Background *background;
+static GFileMonitor *instance_monitor;
 
 GType background_get_type (void) G_GNUC_CONST;
 static void background_iface_init (XdpBackgroundIface *iface);
@@ -241,7 +242,7 @@ handle_request_background_in_thread_func (GTask *task,
   else
     permission = get_permission (app_id);
 
-  g_debug ("Handle RequestBackground for %s\n", app_id);
+  g_debug ("Handle RequestBackground for '%s'", app_id);
 
   if (permission == PERMISSION_ASK || permission == PERMISSION_UNSET)
     {
@@ -359,6 +360,13 @@ validate_commandline (const char *key,
   gsize length;
   const char **strv = g_variant_get_strv (value, &length);
 
+  if (strv[0] == NULL)
+    {
+      g_set_error (error, XDG_DESKTOP_PORTAL_ERROR, XDG_DESKTOP_PORTAL_ERROR_INVALID_ARGUMENT,
+                   "Commandline can't be empty");
+      return FALSE;
+    }
+
   if (g_utf8_strlen (strv[0], -1) > 256)
     {
       g_set_error (error, XDG_DESKTOP_PORTAL_ERROR, XDG_DESKTOP_PORTAL_ERROR_INVALID_ARGUMENT,
@@ -398,9 +406,13 @@ handle_request_background (XdpBackground *object,
   REQUEST_AUTOLOCK (request);
 
   g_variant_builder_init (&opt_builder, G_VARIANT_TYPE_VARDICT);
-  xdp_filter_options (arg_options, &opt_builder,
-                      background_options, G_N_ELEMENTS (background_options),
-                      NULL);
+  if (!xdp_filter_options (arg_options, &opt_builder,
+                           background_options, G_N_ELEMENTS (background_options),
+                           &error))
+    {
+      g_dbus_method_invocation_return_gerror (invocation, error);
+      return TRUE;
+    }
 
   options = g_variant_ref_sink (g_variant_builder_end (&opt_builder));
 
@@ -449,6 +461,22 @@ background_class_init (BackgroundClass *klass)
 
 /* background monitor */
 
+/* The background monitor is running in a dedicated thread.
+ *
+ * We rely on the RunningApplicationsChanged signal from the backend to get
+ * notified about applications that start or stop having open windows, and on
+ * file monitoring to learn about flatpak instances appearing and disappearing.
+ *
+ * When either of these changes happens, we wake up the background monitor
+ * thread, and it will do a check the state of applications a few times, with
+ * a few seconds of wait in between. When we find an application in the background
+ * more than once, we check the permissions, and kill or notify if warranted.
+ *
+ * We require an application to be in background state for more than once check
+ * to avoid killing an unlucky application that just happend to start up as we
+ * did our check.
+ */
+
 typedef enum { BACKGROUND, RUNNING, ACTIVE } AppState;
 
 static GHashTable *
@@ -462,7 +490,14 @@ get_app_states (void)
 
   if (!xdp_impl_background_call_get_app_state_sync (background_impl, &apps, NULL, &error))
     {
-      g_warning ("Failed to get application states: %s", error->message);
+      static int warned = 0;
+
+      if (!warned)
+        {
+          g_warning ("Failed to get application states: %s", error->message);
+          warned = 1;
+        }
+
       return NULL;
     }
 
@@ -511,7 +546,7 @@ G_LOCK_DEFINE (applications);
 static void
 close_notification (const char *handle)
 {
-  g_dbus_connection_call (g_dbus_proxy_get_connection (G_DBUS_PROXY (background)),
+  g_dbus_connection_call (g_dbus_proxy_get_connection (G_DBUS_PROXY (background_impl)),
                           g_dbus_proxy_get_name (G_DBUS_PROXY (background_impl)),
                           handle,
                           "org.freedesktop.impl.portal.Request",
@@ -623,7 +658,8 @@ notify_background_done (GObject *source,
       return;
     }
 
-  g_variant_lookup (results, "result", "u", &result);
+  if (!g_variant_lookup (results, "result", "u", &result))
+    result = IGNORE;
 
   if (result == ALLOW)
     {
@@ -695,7 +731,6 @@ check_background_apps (void)
       const char *app_id;
       pid_t child_pid;
       InstanceData *idata;
-      Permission permission;
       const char *state_names[] = { "background", "running", "active" };
       gboolean is_new = FALSE;
 
@@ -721,23 +756,22 @@ check_background_apps (void)
 
       g_debug ("App %s is %s", app_id, state_names[idata->state]);
 
-      permission = get_one_permission (app_id, perms);
-
-      if (idata->permission != permission)
-        {
-          /* Notify again if permissions change */
-          idata->permission = permission;
-          idata->notified = FALSE;
-        }
+      idata->permission = get_one_permission (app_id, perms);
 
       /* If the app is not in the list yet, add it,
        * but don't notify yet - this gives apps some
-       * leeway to get their window app. If it is still
+       * leeway to get their window up. If it is still
        * in the background next time around, we'll proceed
        * to the next step.
        */
       if (idata->state != BACKGROUND || idata->notified || is_new)
-        continue;
+        {
+          if (idata->notified)
+            g_debug ("Already notified app %s ...skipping\n", app_id);
+          if (is_new)
+            g_debug ("App %s is new ...skipping\n", app_id);
+          continue;
+        }
 
       switch (idata->permission)
         {
@@ -753,7 +787,12 @@ check_background_apps (void)
           {
             NotificationData *nd = g_new0 (NotificationData, 1);
 
-            g_assert (idata->handle == NULL);
+            if (idata->handle)
+              {
+                close_notification (idata->handle);
+                g_free (idata->handle);
+              }
+
             idata->handle = g_strdup_printf ("/org/freedesktop/portal/desktop/notify/background%d", stamp);
             idata->notified = TRUE;
 
@@ -779,6 +818,10 @@ check_background_apps (void)
     {
       NotificationData *nd = g_ptr_array_index (notifications, i);
 
+      g_debug ("Tentatively allow background for %s", nd->app_id);
+
+      set_permission (nd->app_id, PERMISSION_YES);
+
       g_debug ("Notify background for %s", nd->app_id);
 
       xdp_impl_background_call_notify_background (background_impl,
@@ -793,18 +836,26 @@ check_background_apps (void)
   remove_outdated_instances (stamp);
 }
 
+static GMainContext *monitor_context;
+
 static gpointer
 background_monitor (gpointer data)
 {
   applications = g_hash_table_new_full (g_str_hash, g_str_equal,
                                         g_free, instance_data_free);
-  while (1)
+
+  while (TRUE)
     {
+      g_main_context_iteration (monitor_context, TRUE);
+      /* We check twice, to avoid killing unlucky apps hit at a bad time */
+      sleep (30);
+      check_background_apps ();
+      sleep (30);
       check_background_apps ();
-      sleep (60);
     }
 
   g_clear_pointer (&applications, g_hash_table_unref);
+  g_clear_pointer (&monitor_context, g_main_context_unref);
 
   return NULL;
 }
@@ -816,14 +867,32 @@ start_background_monitor (void)
 
   g_debug ("Starting background app monitor");
 
+  monitor_context = g_main_context_new ();
+
   thread = g_thread_new ("background monitor", background_monitor, NULL);
 }
 
+static void
+running_apps_changed (gpointer data)
+{
+  g_debug ("Running app windows changed, wake up monitor thread");
+  g_main_context_wakeup (monitor_context);
+}
+
+static void
+instances_changed (gpointer data)
+{
+  g_debug ("Running instances changed, wake up monitor thread");
+  g_main_context_wakeup (monitor_context);
+}
+
 GDBusInterfaceSkeleton *
 background_create (GDBusConnection *connection,
                    const char *dbus_name_access,
                    const char *dbus_name_background)
 {
+  g_autofree char *instance_path = NULL;
+  g_autoptr(GFile) instance_dir = NULL;
   g_autoptr(GError) error = NULL;
 
   access_impl = xdp_impl_access_proxy_new_sync (connection,
@@ -857,5 +926,17 @@ background_create (GDBusConnection *connection,
 
   start_background_monitor ();
 
+  g_signal_connect (background_impl, "running-applications-changed",
+                    G_CALLBACK (running_apps_changed), NULL);
+
+  /* FIXME: it would be better if libflatpak had a monitor api for this */
+  instance_path = g_build_filename (g_get_user_runtime_dir (), ".flatpak", NULL);
+  instance_dir = g_file_new_for_path (instance_path);
+  instance_monitor = g_file_monitor_directory (instance_dir, G_FILE_MONITOR_NONE, NULL, &error);
+  if (!instance_monitor)
+    g_warning ("Failed to create a monitor for %s: %s", instance_path, error->message);
+  else
+    g_signal_connect (instance_monitor, "changed", G_CALLBACK (instances_changed), NULL);
+
   return G_DBUS_INTERFACE_SKELETON (background);
 }
