@@ -38,6 +38,7 @@
 #include "file-chooser.h"
 #include "open-uri.h"
 #include "print.h"
+#include "memory-monitor.h"
 #include "network-monitor.h"
 #include "proxy-resolver.h"
 #include "screenshot.h"
@@ -54,6 +55,8 @@
 #include "background.h"
 #include "gamemode.h"
 #include "camera.h"
+#include "secret.h"
+#include "wallpaper.h"
 
 static GMainLoop *loop = NULL;
 
@@ -203,6 +206,11 @@ on_bus_acquired (GDBusConnection *connection,
   PortalImplementation *implementation2;
   g_autoptr(GError) error = NULL;
   XdpImplLockdown *lockdown;
+  GQuark portal_errors G_GNUC_UNUSED;
+  GPtrArray *impls;
+
+  /* make sure errors are registered */
+  portal_errors = XDG_DESKTOP_PORTAL_ERROR;
 
   xdp_connection_track_name_owners (connection, peer_died_cb);
   init_document_proxy (connection);
@@ -218,13 +226,15 @@ on_bus_acquired (GDBusConnection *connection,
   else
     lockdown = xdp_impl_lockdown_skeleton_new ();
 
+  export_portal_implementation (connection, memory_monitor_create (connection));
   export_portal_implementation (connection, network_monitor_create (connection));
   export_portal_implementation (connection, proxy_resolver_create (connection));
   export_portal_implementation (connection, trash_create (connection));
   export_portal_implementation (connection, game_mode_create (connection));
 
-  implementation = find_portal_implementation ("org.freedesktop.impl.portal.Settings");
-  export_portal_implementation (connection, settings_create (connection, implementation ? implementation->dbus_name : NULL));
+  impls = find_all_portal_implementations ("org.freedesktop.impl.portal.Settings");
+  export_portal_implementation (connection, settings_create (connection, impls));
+  g_ptr_array_free (impls, TRUE);
 
   implementation = find_portal_implementation ("org.freedesktop.impl.portal.FileChooser");
   if (implementation != NULL)
@@ -267,8 +277,8 @@ on_bus_acquired (GDBusConnection *connection,
                                     location_create (connection, implementation->dbus_name, lockdown));
 #endif
 
-#if HAVE_PIPEWIRE
-      export_portal_implementation (connection, camera_create (connection));
+#ifdef HAVE_PIPEWIRE
+      export_portal_implementation (connection, camera_create (connection, lockdown));
 #endif
     }
 
@@ -278,6 +288,13 @@ on_bus_acquired (GDBusConnection *connection,
                                                      implementation->dbus_name,
                                                      implementation2->dbus_name));
 
+  implementation2 = find_portal_implementation ("org.freedesktop.impl.portal.Wallpaper");
+  if (implementation != NULL && implementation2 != NULL)
+    export_portal_implementation (connection,
+                                  wallpaper_create (connection,
+                                                    implementation->dbus_name,
+                                                    implementation2->dbus_name));
+
   implementation = find_portal_implementation ("org.freedesktop.impl.portal.Account");
   if (implementation != NULL)
     export_portal_implementation (connection,
@@ -288,7 +305,12 @@ on_bus_acquired (GDBusConnection *connection,
     export_portal_implementation (connection,
                                   email_create (connection, implementation->dbus_name));
 
-#if HAVE_PIPEWIRE
+  implementation = find_portal_implementation ("org.freedesktop.impl.portal.Secret");
+  if (implementation != NULL)
+    export_portal_implementation (connection,
+				  secret_create (connection, implementation->dbus_name));
+
+#ifdef HAVE_PIPEWIRE
   implementation = find_portal_implementation ("org.freedesktop.impl.portal.ScreenCast");
   if (implementation != NULL)
     export_portal_implementation (connection,
@@ -356,7 +378,6 @@ main (int argc, char *argv[])
       g_printerr ("Try \"%s --help\" for more information.",
                   g_get_prgname ());
       g_printerr ("\n");
-      g_option_context_free (context);
       return 1;
     }
 
