@@ -236,7 +236,6 @@ xdp_app_info_rewrite_commandline (XdpAppInfo *app_info,
       if (commandline && commandline[0])
         {
           int i;
-          g_autofree char *cmd = NULL;
 
           g_ptr_array_add (args, g_strdup_printf ("--command=%s", commandline[0]));
           g_ptr_array_add (args, g_strdup (app_info->id));
@@ -441,7 +440,12 @@ parse_app_info_from_flatpak_info (int pid, GError **error)
 
   id = g_key_file_get_string (metadata, group, "name", error);
   if (id == NULL)
-    return NULL;
+    {
+      close (info_fd);
+      return NULL;
+    }
+
+  close (info_fd);
 
   app_info = xdp_app_info_new (XDP_APP_INFO_KIND_FLATPAK);
   app_info->id = g_steal_pointer (&id);
@@ -774,7 +778,21 @@ xdp_filter_options (GVariant *options,
                                       supported_options[i].key,
                                       supported_options[i].type);
       if (!value)
-        continue;
+        {
+          value = g_variant_lookup_value (options, supported_options[i].key, NULL);
+          if (value)
+            {
+              if (*error == NULL)
+                g_set_error (error, XDG_DESKTOP_PORTAL_ERROR, XDG_DESKTOP_PORTAL_ERROR_INVALID_ARGUMENT,
+                             "Expected type '%s' for option '%s', got '%s'",
+                             g_variant_type_peek_string (supported_options[i].type),
+                             supported_options[i].key,
+                             g_variant_type_peek_string (g_variant_get_type (value)));
+              ret = FALSE;
+            }
+
+          continue;
+        }
          
       if (supported_options[i].validate)
         {
@@ -785,8 +803,11 @@ xdp_filter_options (GVariant *options,
               if (ret)
                 {
                   ret = FALSE;
-                  g_propagate_error (error, local_error);
-                  local_error = NULL;
+                  if (error && *error == NULL)
+                    {
+                      g_propagate_error (error, local_error);
+                      local_error = NULL;
+                    }
                 }
 
               continue;
@@ -1437,6 +1458,21 @@ open_pid_fd (int      proc_fd,
   return fd;
 }
 
+static int
+open_fdinfo_dir (GError **error)
+{
+  int fd;
+
+  fd = open ("/proc/self/fdinfo", O_RDONLY | O_NONBLOCK | O_DIRECTORY | O_CLOEXEC | O_NOCTTY);
+
+  if (fd < 0)
+    g_set_error (error, G_IO_ERROR, g_io_error_from_errno (errno),
+                 "Could not to open /proc/self/fdinfo: %s",
+                 g_strerror (errno));
+
+  return fd;
+}
+
 static inline gboolean
 find_pid (pid_t *pids,
           guint  n_pids,
@@ -1468,6 +1504,7 @@ map_pids (DIR     *proc,
   guint count = 0;
 
   res = g_alloca (sizeof (pid_t) * n_pids);
+  memset (res, 0, sizeof (pid_t) * n_pids);
 
   while ((de = readdir (proc)) != NULL)
     {
@@ -1512,14 +1549,31 @@ map_pids (DIR     *proc,
           return FALSE;
         }
 
-      res[idx] = outside;
-      count++;
+      /* this handles the first occurrence, already identified by find_pid,
+       * as well as duplicate entries */
+      for (guint i = idx; i < n_pids; i++)
+        {
+          if (pids[i] == inside)
+            {
+              res[idx] = outside;
+              count++;
+            }
+        }
     }
 
   if (count != n_pids)
     {
-      g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_NOT_FOUND,
-                           "Some process ids could not be found");
+      g_autoptr(GString) str = NULL;
+
+      str = g_string_new ("Process ids could not be found: ");
+
+      for (guint i = 0; i < n_pids; i++)
+        if (res[i] == 0)
+          g_string_append_printf (str, "%d, ", (guint32) pids[i]);
+
+      g_string_truncate (str, str->len - 2);
+      g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_NOT_FOUND, str->str);
+
       return FALSE;
     }
 
@@ -1528,18 +1582,89 @@ map_pids (DIR     *proc,
   return TRUE;
 }
 
-static pid_t
-xdp_app_info_get_child_pid (XdpAppInfo *app_info,
-                            GError    **error)
+static gboolean
+pidfd_to_pid (int fdinfo, const int pidfd, pid_t *pid, GError **error)
+{
+  g_autofree char *name = NULL;
+  g_autofree char *key = NULL;
+  g_autofree char *val = NULL;
+  gboolean found = FALSE;
+  FILE *f = NULL;
+  size_t keylen = 0;
+  size_t vallen = 0;
+  ssize_t n;
+  int fd;
+  int r = 0;
+
+  *pid = 0;
+
+  name = g_strdup_printf ("%d", pidfd);
+
+  fd = openat (fdinfo, name, O_RDONLY | O_CLOEXEC | O_NOCTTY);
+
+  if (fd != -1)
+    f = fdopen (fd, "r");
+
+  if (f == NULL)
+    {
+      g_set_error (error, G_IO_ERROR, g_io_error_from_errno (errno),
+                   "Unable to open /proc/self/fdinfo/%d: %s",
+                   fd, g_strerror (errno));
+      return FALSE;
+    }
+
+  do {
+    n = getdelim (&key, &keylen, ':', f);
+    if (n == -1)
+      {
+        r = errno;
+        break;
+      }
+
+    n = getdelim (&val, &vallen, '\n', f);
+    if (n == -1)
+      {
+        r = errno;
+        break;
+      }
+
+    g_strstrip (key);
+
+    if (!strncmp (key, "Pid", 3))
+      {
+        r = parse_status_field_pid (val, pid);
+        found = r > -1;
+      }
+
+  } while (r == 0 && !found);
+
+  fclose (f);
+
+  if (r < 0)
+    {
+      g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
+                   "Could not parse fdinfo::%s: %s",
+                   key, g_strerror (-r));
+    }
+  else if (!found)
+    {
+      g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_FAILED,
+                           "Could not parse fdinfo: Pid field missing");
+    }
+
+  return found;
+}
+
+static JsonNode *
+xdg_app_info_load_brwap_info (XdpAppInfo *app_info,
+                              GError    **error)
 {
   g_autoptr(JsonParser) parser = NULL;
+  g_autoptr(JsonNode) root = NULL;
   g_autofree char *instance = NULL;
   g_autofree char *data = NULL;
-  JsonNode *root;
-  JsonObject *cpo;
   gsize len;
   char *path;
-  pid_t pid;
 
   g_return_val_if_fail (app_info != NULL, 0);
 
@@ -1564,7 +1689,7 @@ xdp_app_info_get_child_pid (XdpAppInfo *app_info,
       return 0;
     }
 
-  root = json_parser_get_root (parser);
+  root = json_parser_steal_root (parser);
   if (!root)
     {
       g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
@@ -1579,12 +1704,45 @@ xdp_app_info_get_child_pid (XdpAppInfo *app_info,
       return 0;
     }
 
+  return g_steal_pointer (&root);
+}
+
+static ino_t
+xdp_app_info_get_pid_namespace (JsonNode *root,
+                                GError  **error)
+{
+  JsonNode *node;
+  JsonObject *cpo;
+  gint64 nsid;
+
+  /* xdg_app_info_load_brwap_info assures root is of type object */
+  cpo = json_node_get_object (root);
+  node = json_object_get_member (cpo, "pid-namespace");
+
+  if (node == NULL || !JSON_NODE_HOLDS_VALUE (node))
+    {
+      g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_NOT_FOUND,
+                           "pid-namespace missing");
+      return 0;
+    }
+
+  nsid = json_node_get_int (node);
+  return (ino_t) nsid;
+}
+
+static pid_t
+xdp_app_info_get_child_pid (JsonNode *root,
+                            GError  **error)
+{
+  JsonObject *cpo;
+  pid_t pid;
+
   cpo = json_node_get_object (root);
 
   pid = json_object_get_int_member (cpo, "child-pid");
   if (pid == 0)
-    g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
-                 "Could not parse '%s': child-pid missing", path);
+    g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_NOT_FOUND,
+                         "child-pid missing");
 
   return pid;
 }
@@ -1596,6 +1754,7 @@ xdg_app_info_ensure_pidns (XdpAppInfo  *app_info,
                            DIR         *proc,
                            GError     **error)
 {
+  g_autoptr(JsonNode) root = NULL;
   xdp_lockguard GMutex *guard = NULL;
   xdp_autofd int fd = -1;
   pid_t pid;
@@ -1607,7 +1766,22 @@ xdg_app_info_ensure_pidns (XdpAppInfo  *app_info,
   if (app_info->u.flatpak.pidns_id != 0)
     return TRUE;
 
-  pid = xdp_app_info_get_child_pid (app_info, error);
+  root = xdg_app_info_load_brwap_info (app_info, error);
+  if (root == NULL)
+    return FALSE;
+
+  /* newer versions of bubblewrap contain the namespace
+   * information directly, so we don' thave to go via the
+   * child-pid; if this fails, we fallback to the old way */
+  ns = xdp_app_info_get_pid_namespace (root, NULL);
+  if (ns != 0)
+    {
+      g_debug ("Using pid namespace info from bwrap info");
+      app_info->u.flatpak.pidns_id = ns;
+      return TRUE;
+    }
+
+  pid = xdp_app_info_get_child_pid (root, error);
   if (pid == 0)
     return FALSE;
 
@@ -1630,6 +1804,8 @@ xdg_app_info_ensure_pidns (XdpAppInfo  *app_info,
   return TRUE;
 }
 
+
+
 gboolean
 xdg_app_info_map_pids (XdpAppInfo  *app_info,
                        pid_t       *pids,
@@ -1679,3 +1855,29 @@ xdg_app_info_map_pids (XdpAppInfo  *app_info,
   closedir (proc);
   return ok;
 }
+
+gboolean
+xdg_app_info_pidfds_to_pids (XdpAppInfo  *app_info,
+                             const int   *fds,
+                             pid_t       *pids,
+                             gint         count,
+                             GError     **error)
+{
+  gboolean ok = TRUE;
+  int fdinfo = -1;
+
+  g_return_val_if_fail (app_info != NULL, FALSE);
+  g_return_val_if_fail (fds != NULL, FALSE);
+  g_return_val_if_fail (pids != NULL, FALSE);
+
+  fdinfo = open_fdinfo_dir (error);
+  if (fdinfo == -1)
+    return FALSE;
+
+  for (gint i = 0; i < count && ok; i++)
+    ok = pidfd_to_pid (fdinfo, fds[i], &pids[i], error);
+
+  (void) close (fdinfo);
+
+  return ok;
+}
